"""Base ThermoState module."""
from math import isclose
import sys

import CoolProp
from pint import UnitRegistry, DimensionalityError
from pint.unit import UnitsContainer, UnitDefinition
from pint.converters import ScaleConverter

try:  # pragma: no cover
    from IPython.core.ultratb import AutoFormattedTB
except ImportError:  # pragma: no cover
    AutoFormattedTB = None

units = UnitRegistry(autoconvert_offset_to_baseunit=True)
Q_ = units.Quantity
units.define(UnitDefinition("percent", "pct", (), ScaleConverter(1.0 / 100.0)))

# Don't add the _render_traceback_ function to DimensionalityError if
# IPython isn't present. This function is only used by the IPython/ipykernel
# anyways, so it doesn't matter if it's missing if IPython isn't available.
if AutoFormattedTB is not None:  # pragma: no cover

    def render_traceback(self):
        """Render a minimized version of the DimensionalityError traceback.

        The default Jupyter/IPython traceback includes a lot of
        context from within pint that actually raises the
        DimensionalityError. This context isn't really needed for
        this particular error, since the problem is almost certainly in
        the user code. This function removes the additional context.
        """
        a = AutoFormattedTB(mode="Context", color_scheme="Neutral", tb_offset=1)
        etype, evalue, tb = sys.exc_info()
        stb = a.structured_traceback(etype, evalue, tb, tb_offset=1)
        for i, line in enumerate(stb):
            if "site-packages" in line:
                first_line = i
                break
        return stb[:first_line] + stb[-1:]

    DimensionalityError._render_traceback_ = render_traceback.__get__(
        DimensionalityError
    )

phase_map = {
    getattr(CoolProp, i): i.split("_")[1] for i in dir(CoolProp) if "iphase" in i
}


def munge_coolprop_input_prop(prop):
    prop = prop.replace("_INPUTS", "").replace("mass", "").replace("D", "V")
    return prop.replace("Q", "X").lower().replace("t", "T")


def isclose_quant(a, b, *args, **kwargs):
    return isclose(a.magnitude, b.magnitude, *args, **kwargs)


class StateError(Exception):
    """Errors associated with setting the `State` object."""

    def _render_traceback_(self):  # pragma: no cover
        """Render a minimized version of the `StateError` traceback.

        The default Jupyter/IPython traceback includes a lot of
        context from within `State` where the `StateError` is raised.
        This context isn't really needed, since the problem is almost certainly in
        the user code. This function removes the additional context.
        """
        if AutoFormattedTB is not None:
            a = AutoFormattedTB(mode="Context", color_scheme="Neutral", tb_offset=1)
            etype, evalue, tb = sys.exc_info()
            stb = a.structured_traceback(etype, evalue, tb, tb_offset=1)
            for i, line in enumerate(stb):
                if "site-packages" in line:
                    first_line = i
                    break
            return stb[:first_line] + stb[-1:]


class State(object):
    """Basic State manager for thermodyanmic states.

    Parameters
    ----------
    substance : `str`
        One of the substances supported by CoolProp
    T : `pint.UnitRegistry.Quantity`
        Temperature
    p : `pint.UnitRegistry.Quantity`
        Pressure
    u : `pint.UnitRegistry.Quantity`
        Mass-specific internal energy
    s : `pint.UnitRegistry.Quantity`
        Mass-specific entropy
    v : `pint.UnitRegistry.Quantity`
        Mass-specific volume
    h : `pint.UnitRegistry.Quantity`
        Mass-specific enthalpy
    x : `pint.UnitRegistry.Quantity`
        Quality

    """

    _allowed_subs = [
        "AIR",
        "AMMONIA",
        "WATER",
        "PROPANE",
        "R134A",
        "R22",
        "ISOBUTANE",
        "CARBONDIOXIDE",
        "OXYGEN",
        "NITROGEN",
    ]

    _all_pairs = [
        munge_coolprop_input_prop(k)
        for k in dir(CoolProp.constants)
        if "INPUTS" in k and "molar" not in k
    ]
    _all_pairs.extend([k[::-1] for k in _all_pairs])

    _unsupported_pairs = ["Tu", "Th", "us"]
    _unsupported_pairs.extend([k[::-1] for k in _unsupported_pairs])

    # This weird lambda construct is necessary because _unsupported_pairs can't be accessed
    # inside the list comprehension because of the namespacing rules for class attributes.
    # Trying to set _allowed_pairs in the __init__ leads to infinite recursion because of
    # how we're messing with __setattr__.
    _allowed_pairs = (
        lambda x=_unsupported_pairs, y=_all_pairs: [p for p in y if p not in x]
    )()

    _all_props = list("Tpvuhsx")

    _read_only_props = ["cp", "cv", "phase"]

    _dimensions = {
        "T": UnitsContainer({"[temperature]": 1.0}),
        "p": UnitsContainer({"[mass]": 1.0, "[length]": -1.0, "[time]": -2.0}),
        "v": UnitsContainer({"[length]": 3.0, "[mass]": -1.0}),
        "u": UnitsContainer({"[length]": 2.0, "[time]": -2.0}),
        "h": UnitsContainer({"[length]": 2.0, "[time]": -2.0}),
        "s": UnitsContainer({"[length]": 2.0, "[time]": -2.0, "[temperature]": -1.0}),
        "x": UnitsContainer({}),
    }

    _SI_units = {
        "T": "kelvin",
        "p": "pascal",
        "v": "meter**3/kilogram",
        "u": "joules/kilogram",
        "h": "joules/kilogram",
        "s": "joules/(kilogram*kelvin)",
        "x": "dimensionless",
        "cp": "joules/(kilogram*kelvin)",
        "cv": "joules/(kilogram*kelvin)",
        "phase": None,
    }

    def __setattr__(self, key, value):
        if key.startswith("_") or key == "sub":
            object.__setattr__(self, key, value)
        elif key in self._allowed_pairs:
            self._check_dimensions(key, value)
            self._check_values(key, value)
            self._set_properties(key, value)
        elif key in self._unsupported_pairs:
            raise StateError(
                "The pair of input properties entered ({}) isn't supported yet. "
                "Sorry!".format(key)
            )
        else:
            raise AttributeError("Unknown attribute {}".format(key))

    def __getattr__(self, key):
        if key in self._all_props:
            return object.__getattribute__(self, "_" + key)
        elif key in self._all_pairs:
            val_0 = object.__getattribute__(self, "_" + key[0])
            val_1 = object.__getattribute__(self, "_" + key[1])
            return val_0, val_1
        elif key == "phase":
            return object.__getattribute__(self, "_" + key)
        elif key in self._read_only_props:
            return object.__getattribute__(self, "_" + key)
        else:
            raise AttributeError("Unknown attribute {}".format(key))

    def __eq__(self, other):
        """Check if two `State`s are equivalent.

        Check that they are using the same substance and two properties that are
        always independent. Choose T and v because the EOS tends to be defined
        in terms of T and density.
        """
        if (
            self.sub == other.sub
            and isclose_quant(other.T, self.T)
            and isclose_quant(other.v, self.v)
        ):
            return True

    def __le__(self, other):
        return NotImplemented

    def __lt__(self, other):
        return NotImplemented

    def __gt__(self, other):
        return NotImplemented

    def __ge__(self, other):
        return NotImplemented

    def __init__(self, substance, **kwargs):
        if substance.upper() in self._allowed_subs:
            self.sub = substance.upper()
        else:
            raise ValueError(
                "{} is not an allowed substance. Choose one of {}.".format(
                    substance, self._allowed_subs
                )
            )

        self._abstract_state = CoolProp.AbstractState("HEOS", self.sub)

        input_props = ""
        for arg in kwargs:
            if arg not in self._all_props:
                raise ValueError("The argument {} is not allowed.".format(arg))
            else:
                input_props += arg

        if len(input_props) > 2 or len(input_props) == 1:
            raise ValueError(
                "Incorrect number of properties specified. Must be 2 or 0."
            )

        if len(input_props) > 0 and input_props not in self._allowed_pairs:
            raise StateError(
                "The pair of input properties entered ({}) isn't supported yet. "
                "Sorry!".format(input_props)
            )

        if len(input_props) > 0:
            setattr(self, input_props, (kwargs[input_props[0]], kwargs[input_props[1]]))

    def to_SI(self, prop, value):
        return value.to(self._SI_units[prop])

    def to_PropsSI(self, prop, value):
        return self.to_SI(prop, value).magnitude

    def _check_values(self, properties, value):
        for i in range(len(properties)):
            if properties[i] in "Tvp" and value[i].to_base_units().magnitude < 0.0:
                raise StateError(
                    "The value of {} must be positive in absolute units".format(
                        properties[i]
                    )
                )
            elif properties[i] in "x" and (
                value[i].to_base_units().magnitude < 0.0
                or value[i].to_base_units().magnitude > 1.0
            ):
                raise StateError("The value of the quality must be between 0 and 1")

    def _check_dimensions(self, properties, value):
        if value[0].dimensionality != self._dimensions[properties[0]]:
            raise StateError(
                "The dimensions for {props[0]} must be {dim}".format(
                    props=properties, dim=self._dimensions[properties[0]]
                )
            )
        elif value[1].dimensionality != self._dimensions[properties[1]]:
            raise StateError(
                "The dimensions for {props[1]} must be {dim}".format(
                    props=properties, dim=self._dimensions[properties[1]]
                )
            )

    def _set_properties(self, known_props, known_values):
        if (
            len(known_props) != 2
            or len(known_values) != 2
            or len(known_props) != len(known_values)
        ):
            raise StateError("Only specify two properties to _set_properties")

        known_state = []

        for prop, val in zip(known_props, known_values):
            if prop == "x":
                known_state.append(("Q", self.to_PropsSI(prop, val)))
            elif prop == "v":
                known_state.append(("Dmass", 1.0 / self.to_PropsSI(prop, val)))
            else:
                postfix = "" if prop in ["T", "p"] else "mass"
                known_state.append((prop.upper() + postfix, self.to_PropsSI(prop, val)))

        known_state.sort(key=lambda p: p[0])

        inputs = getattr(CoolProp, "".join([p[0] for p in known_state]) + "_INPUTS")
        try:
            self._abstract_state.update(inputs, known_state[0][1], known_state[1][1])
        except ValueError as e:
            if "Saturation pressure" in str(e):
                raise StateError(
                    "The given values for {} and {} are not "
                    "independent.".format(known_props[0], known_props[1])
                )
            else:
                raise

        for prop in self._all_props + self._read_only_props:
            if prop == "v":
                value = Q_(
                    1.0 / self._abstract_state.keyed_output(CoolProp.iDmass),
                    self._SI_units[prop],
                )
            elif prop == "x":
                value = Q_(
                    self._abstract_state.keyed_output(CoolProp.iQ), self._SI_units[prop]
                )
                if value == -1.0:
                    value = None
            elif prop == "phase":
                value = phase_map[self._abstract_state.keyed_output(CoolProp.iPhase)]
            else:
                postfix = "" if prop in ["T", "p"] else "mass"
                p = getattr(CoolProp, "i" + prop.title() + postfix)
                value = Q_(self._abstract_state.keyed_output(p), self._SI_units[prop])

            setattr(self, "_" + prop, value)



"""
Test module for the units abbreviations code
"""
from ..abbreviations import EnglishEngineering as EE, SystemInternational as SI


def test_EE():
    assert EE.s == 'BTU/(lb*degR)'
    assert EE.h == 'BTU/lb'
    assert EE.T == 'degF'
    assert EE.u == 'BTU/lb'
    assert EE.v == 'ft**3/lb'
    assert EE.p == 'psi'


def test_SI():
    assert SI.s == 'kJ/(kg*K)'
    assert SI.h == 'kJ/kg'
    assert SI.T == 'degC'
    assert SI.u == 'kJ/kg'
    assert SI.v == 'm**3/kg'
    assert SI.p == 'bar'



"""
This module contains classes with attributes representing the common property units
"""


class EnglishEngineering(object):
    """String representations of common units.

    The attributes of this class are strings that represent the common units for
    thermodynamics calculations.

    Attributes
    ----------
    h : `str`
        BTU/lb
    p : `str`
        psi
    s : `str`
        BTU/(lb*degR)
    T : `str`
        degF
    u : `str`
        BTU/lb
    v : `str`
        ft**3/lb
    """

    s = 'BTU/(lb*degR)'
    h = 'BTU/lb'
    T = 'degF'
    u = 'BTU/lb'
    v = 'ft**3/lb'
    p = 'psi'


class SystemInternational(object):
    """String representations of common units.

    The attributes of this class are strings that represent the common units for
    thermodynamics calculations.

    Attributes
    ----------
    h : `str`
        kJ/kg
    p : `str`
        bar
    s : `str`
        kJ/(kg*K)
    T : `str`
        degC
    u : `str`
        kJ/kg
    v : `str`
        m**3/lb
    """

    s = 'kJ/(kg*K)'
    h = 'kJ/kg'
    T = 'degC'
    u = 'kJ/kg'
    v = 'm**3/kg'
    p = 'bar'



"""
Test module for the main ThermoState code
"""
import pytest

from ..thermostate import State, StateError, Q_, isclose_quant


class TestState(object):
    """
    """
    def test_eq(self):
        st_1 = State(substance='water', T=Q_(400.0, 'K'), p=Q_(101325.0, 'Pa'))
        st_2 = State(substance='water', T=Q_(400.0, 'K'), p=Q_(101325.0, 'Pa'))
        assert st_1 == st_2

    def test_not_eq(self):
        st_1 = State(substance='water', T=Q_(400.0, 'K'), p=Q_(101325.0, 'Pa'))
        st_2 = State(substance='water', T=Q_(300.0, 'K'), p=Q_(101325.0, 'Pa'))
        assert not st_1 == st_2

    def test_not_eq_sub(self):
        st_1 = State(substance='water', T=Q_(400.0, 'K'), p=Q_(101325.0, 'Pa'))
        st_2 = State(substance='ammonia', T=Q_(400.0, 'K'), p=Q_(101325.0, 'Pa'))
        assert not st_1 == st_2

    def test_comparison(self):
        st_1 = State(substance='water', T=Q_(400.0, 'K'), p=Q_(101325.0, 'Pa'))
        st_2 = State(substance='water', T=Q_(400.0, 'K'), p=Q_(101325.0, 'Pa'))
        with pytest.raises(TypeError):
            st_1 < st_2
        with pytest.raises(TypeError):
            st_1 <= st_2
        with pytest.raises(TypeError):
            st_1 > st_2
        with pytest.raises(TypeError):
            st_1 >= st_2

    def test_unit_definitions(self):
        st = State('water')
        l = st._all_props[:]
        l.extend(st._read_only_props)
        assert all([a in st._SI_units.keys() for a in l])

    def test_lowercase_input(self):
        State(substance='water')
        State(substance='r22')
        State(substance='r134a')
        State(substance='ammonia')
        State(substance='propane')
        State(substance='air')
        State(substance='isobutane')
        State(substance='carbondioxide')
        State(substance='oxygen')
        State(substance='nitrogen')

    def test_bad_substance(self):
        with pytest.raises(ValueError):
            State(substance='bad substance')

    def test_too_many_props(self):
        with pytest.raises(ValueError):
            State(substance='water', T=Q_(300, 'K'), p=Q_(101325, 'Pa'), u=Q_(100, 'kJ/kg'))

    def test_too_few_props(self):
        with pytest.raises(ValueError):
            State(substance='water', T=Q_(300, 'K'))

    def test_negative_temperature(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(-100, "K"), p=Q_(101325, "Pa"))

    def test_negative_pressure(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(300, "K"), p=Q_(-101325, "Pa"))

    def test_negative_volume(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(300, "K"), v=Q_(-10.13, "m**3/kg"))

    def test_quality_lt_zero(self):
        with pytest.raises(StateError):
            State(substance='water', x=Q_(-1.0, "dimensionless"), p=Q_(101325, "Pa"))

    def test_quality_gt_one(self):
        with pytest.raises(StateError):
            State(substance='water', x=Q_(2.0, "dimensionless"), p=Q_(101325, "Pa"))

    def test_invalid_pair(self):
        with pytest.raises(ValueError):
            State(substance='water', x=Q_(0.5, 'dimensionless'), h=Q_(300, 'kJ/kg'))

    def test_invalid_input_prop(self):
        with pytest.raises(ValueError):
            State(substance='water', x=Q_(0.5, 'dimensionless'), bad_prop=Q_(101325, 'Pa'))

    def test_bad_T_dimensions(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(300., 'Pa'), p=Q_(101325., 'Pa'))

    def test_bad_p_dimensions(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(300., 'K'), p=Q_(101325., 'K'))

    def test_bad_u_dimensions(self):
        with pytest.raises(StateError):
            State(substance='water', v=Q_(3., 'm**3/kg'), u=Q_(101325., 'K'))

    def test_bad_s_dimensions(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(300., 'K'), s=Q_(10.1325, 'K'))

    def test_bad_v_dimensions(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(300., 'K'), v=Q_(1.01325, 'K'))

    def test_bad_h_dimensions(self):
        with pytest.raises(StateError):
            State(substance='water', h=Q_(300., 'K'), v=Q_(1.01325, 'K'))

    def test_bad_x_dimensions(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(300., 'K'), x=Q_(1.01325, 'K'))

    def test_TP_twophase(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(373.1242958476844, 'K'), p=Q_(101325., 'Pa'))

    def test_bad_set_properties(self):
        s = State(substance='water')
        with pytest.raises(StateError):
            s._set_properties(['T', 'v', 'h'], [])

    def test_bad_get_property(self):
        s = State(substance='water', T=Q_(400., 'K'), p=Q_(101325., 'Pa'))
        with pytest.raises(AttributeError):
            s.bad_get

    def test_bad_property_setting(self):
        s = State(substance='water')
        with pytest.raises(AttributeError):
            # Should be lowercase p
            s.TP = Q_(400., 'K'), Q_(101325., 'Pa')

    def test_unallowed_pair(self):
        with pytest.raises(StateError):
            State(substance='water', T=Q_(400., 'K'), u=Q_(2547715.3635084038, 'J/kg'))

    def test_set_Tp(self):
        s = State(substance='water')
        s.Tp = Q_(400., 'K'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.Tp[0], Q_(400., 'K'))
        assert isclose_quant(s.Tp[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None
        assert s.phase == 'gas'

    def test_set_pT(self):
        s = State(substance='water')
        s.pT = Q_(101325., 'Pa'), Q_(400., 'K')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.pT[1], Q_(400., 'K'))
        assert isclose_quant(s.pT[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None
        assert s.phase == 'gas'

    # This set of tests fails because T and u are not valid inputs for PhaseSI
    # in CoolProp 6.1.0
    @pytest.mark.xfail(strict=True, raises=StateError)
    def test_set_uT(self):
        s = State(substance='water')
        s.uT = Q_(2547715.3635084038, 'J/kg'), Q_(400., 'K')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.uT[1], Q_(400., 'K'))
        assert isclose_quant(s.uT[0], Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None

    @pytest.mark.xfail(strict=True, raises=StateError)
    def test_set_Tu(self):
        s = State(substance='water')
        s.Tu = Q_(400., 'K'), Q_(2547715.3635084038, 'J/kg')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.Tu[0], Q_(400., 'K'))
        assert isclose_quant(s.Tu[1], Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None

    # This set of tests fails because T and h are not valid inputs for PhaseSI
    # in CoolProp 6.1.0
    @pytest.mark.xfail(strict=True, raises=StateError)
    def test_set_hT(self):
        s = State(substance='water')
        s.hT = Q_(2730301.3859201893, 'J/kg'), Q_(400., 'K')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.hT[1], Q_(400., 'K'))
        assert isclose_quant(s.hT[0], Q_(2730301.3859201893, 'J/kg'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None

    @pytest.mark.xfail(strict=True, raises=StateError)
    def test_set_Th(self):
        s = State(substance='water')
        s.Th = Q_(400., 'K'), Q_(2730301.3859201893, 'J/kg')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.Th[0], Q_(400., 'K'))
        assert isclose_quant(s.Th[1], Q_(2730301.3859201893, 'J/kg'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None

    def test_set_sT(self):
        s = State(substance='water')
        s.sT = Q_(7496.2021523754065, 'J/(kg*K)'), Q_(400., 'K')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.sT[1], Q_(400., 'K'))
        assert isclose_quant(s.sT[0], Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None

    def test_set_Ts(self):
        s = State(substance='water')
        s.Ts = Q_(400., 'K'), Q_(7496.2021523754065, 'J/(kg*K)')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.Ts[0], Q_(400., 'K'))
        assert isclose_quant(s.Ts[1], Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None

    def test_set_vT(self):
        s = State(substance='water')
        s.vT = Q_(1.801983936953226, 'm**3/kg'), Q_(400., 'K')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.vT[1], Q_(400., 'K'))
        assert isclose_quant(s.vT[0], Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None

    def test_set_Tv(self):
        s = State(substance='water')
        s.Tv = Q_(400., 'K'), Q_(1.801983936953226, 'm**3/kg')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.Tv[0], Q_(400., 'K'))
        assert isclose_quant(s.Tv[1], Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.u, Q_(2547715.3635084038, 'J/kg'))
        assert isclose_quant(s.s, Q_(7496.2021523754065, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(2009.2902478486988, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(1509.1482452129906, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(1.801983936953226, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(2730301.3859201893, 'J/kg'))
        assert s.x is None

    def test_set_xT(self):
        s = State(substance='water')
        s.xT = Q_(0.5, 'dimensionless'), Q_(400., 'K')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(245769.34557103913, 'Pa'))
        assert isclose_quant(s.xT[1], Q_(400., 'K'))
        assert isclose_quant(s.xT[0], Q_(0.5, 'dimensionless'))
        assert isclose_quant(s.u, Q_(1534461.5163075812, 'J/kg'))
        assert isclose_quant(s.s, Q_(4329.703956664546, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(4056.471547685226, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(2913.7307270395363, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.3656547423394701, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1624328.2430353598, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.5, 'dimensionless'))
        s.xT = Q_(50, 'percent'), Q_(400., 'K')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(245769.34557103913, 'Pa'))
        assert isclose_quant(s.xT[1], Q_(400., 'K'))
        assert isclose_quant(s.xT[0], Q_(0.5, 'dimensionless'))
        assert isclose_quant(s.u, Q_(1534461.5163075812, 'J/kg'))
        assert isclose_quant(s.s, Q_(4329.703956664546, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(4056.471547685226, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(2913.7307270395363, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.3656547423394701, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1624328.2430353598, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.5, 'dimensionless'))

    def test_set_Tx(self):
        s = State(substance='water')
        s.Tx = Q_(400., 'K'), Q_(0.5, 'dimensionless')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(245769.34557103913, 'Pa'))
        assert isclose_quant(s.Tx[0], Q_(400., 'K'))
        assert isclose_quant(s.Tx[1], Q_(0.5, 'dimensionless'))
        assert isclose_quant(s.u, Q_(1534461.5163075812, 'J/kg'))
        assert isclose_quant(s.s, Q_(4329.703956664546, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(4056.471547685226, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(2913.7307270395363, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.3656547423394701, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1624328.2430353598, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.5, 'dimensionless'))
        s.Tx = Q_(400., 'K'), Q_(50, 'percent')
        assert isclose_quant(s.T, Q_(400., 'K'))
        assert isclose_quant(s.p, Q_(245769.34557103913, 'Pa'))
        assert isclose_quant(s.Tx[0], Q_(400., 'K'))
        assert isclose_quant(s.Tx[1], Q_(0.5, 'dimensionless'))
        assert isclose_quant(s.u, Q_(1534461.5163075812, 'J/kg'))
        assert isclose_quant(s.s, Q_(4329.703956664546, 'J/(kg*K)'))
        assert isclose_quant(s.cp, Q_(4056.471547685226, 'J/(kg*K)'))
        assert isclose_quant(s.cv, Q_(2913.7307270395363, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.3656547423394701, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1624328.2430353598, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.5, 'dimensionless'))

    def test_set_pu(self):
        s = State(substance='water')
        s.pu = Q_(101325., 'Pa'), Q_(1013250., 'J/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.pu[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.pu[1], Q_(1013250., 'J/kg'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))
        s.pu = Q_(101325., 'Pa'), Q_(3013250., 'J/kg')
        assert isclose_quant(s.T, Q_(700.9882316847855, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.pu[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.pu[1], Q_(3013250., 'J/kg'))
        assert isclose_quant(s.u, Q_(3013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(3336406.139862406, 'J/kg'))
        assert s.x is None

    def test_set_up(self):
        s = State(substance='water')
        s.up = Q_(1013250., 'J/kg'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.up[0], Q_(1013250., 'J/kg'))
        assert isclose_quant(s.up[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))
        s.up = Q_(3013250., 'J/kg'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(700.9882316847855, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.up[0], Q_(3013250., 'J/kg'))
        assert isclose_quant(s.up[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(3013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(3336406.139862406, 'J/kg'))
        assert s.x is None

    def test_set_ps(self):
        s = State(substance='water')
        s.ps = Q_(101325., 'Pa'), Q_(3028.9867985920914, 'J/(kg*K)')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.ps[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.ps[1], Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))
        s.ps = Q_(101325., 'Pa'), Q_(8623.283568815832, 'J/(kg*K)')
        assert isclose_quant(s.T, Q_(700.9882316847855, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.ps[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.ps[1], Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.u, Q_(3013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(3336406.139862406, 'J/kg'))
        assert s.x is None

    def test_set_sp(self):
        s = State(substance='water')
        s.sp = Q_(3028.9867985920914, 'J/(kg*K)'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.sp[0], Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.sp[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))
        s.sp = Q_(8623.283568815832, 'J/(kg*K)'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(700.9882316847855, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.sp[0], Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.sp[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(3013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(3336406.139862406, 'J/kg'))
        assert s.x is None

    def test_set_pv(self):
        s = State(substance='water')
        s.pv = Q_(101325., 'Pa'), Q_(0.4772010021515822, 'm**3/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.pv[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.pv[1], Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))
        s.pv = Q_(101325., 'Pa'), Q_(3.189303132125469, 'm**3/kg')
        assert isclose_quant(s.T, Q_(700.9882316847855, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.pv[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.pv[1], Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.u, Q_(3013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(3336406.139862406, 'J/kg'))
        assert s.x is None

    def test_set_vp(self):
        s = State(substance='water')
        s.vp = Q_(0.4772010021515822, 'm**3/kg'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.vp[0], Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.vp[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))
        s.vp = Q_(3.189303132125469, 'm**3/kg'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(700.9882316847855, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.vp[0], Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.vp[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(3013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(3336406.139862406, 'J/kg'))
        assert s.x is None

    def test_set_ph(self):
        s = State(substance='water')
        s.ph = Q_(101325., 'Pa'), Q_(1061602.391543017, 'J/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.ph[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.ph[1], Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))
        s.ph = Q_(101325., 'Pa'), Q_(3336406.139862406, 'J/kg')
        assert isclose_quant(s.T, Q_(700.9882316847855, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.ph[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.ph[1], Q_(3336406.139862406, 'J/kg'))
        assert isclose_quant(s.u, Q_(3013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(3336406.139862406, 'J/kg'))
        assert s.x is None

    def test_set_hp(self):
        s = State(substance='water')
        s.hp = Q_(1061602.391543017, 'J/kg'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.hp[0], Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.hp[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))
        s.hp = Q_(3336406.139862406, 'J/kg'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(700.9882316847855, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.hp[0], Q_(3336406.139862406, 'J/kg'))
        assert isclose_quant(s.hp[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(3013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(8623.283568815832, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(3.189303132125469, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(3336406.139862406, 'J/kg'))
        assert s.x is None

    def test_set_px(self):
        s = State(substance='water')
        s.px = Q_(101325., 'Pa'), Q_(0.28475636946248034, 'dimensionless')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.px[0], Q_(101325., 'Pa'))
        assert isclose_quant(s.px[1], Q_(0.28475636946248034, 'dimensionless'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_xp(self):
        s = State(substance='water')
        s.xp = Q_(0.28475636946248034, 'dimensionless'), Q_(101325., 'Pa')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.xp[0], Q_(0.28475636946248034, 'dimensionless'))
        assert isclose_quant(s.xp[1], Q_(101325., 'Pa'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    # This set of tests fails because s and u are not valid inputs for PhaseSI
    # in CoolProp 6.1.0
    @pytest.mark.xfail(strict=True, raises=StateError)
    def test_set_us(self):
        s = State(substance='water')
        s.us = Q_(1013250., 'J/kg'), Q_(3028.9867985920914, 'J/(kg*K)')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.us[0], Q_(1013250., 'J/kg'))
        assert isclose_quant(s.us[1], Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    @pytest.mark.xfail(strict=True, raises=StateError)
    def test_set_su(self):
        s = State(substance='water')
        s.su = Q_(3028.9867985920914, 'J/(kg*K)'), Q_(1013250., 'J/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.su[0], Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.su[1], Q_(1013250., 'J/kg'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_uv(self):
        s = State(substance='water')
        s.uv = Q_(1013250., 'J/kg'), Q_(0.4772010021515822, 'm**3/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.uv[0], Q_(1013250., 'J/kg'))
        assert isclose_quant(s.uv[1], Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_vu(self):
        s = State(substance='water')
        s.vu = Q_(0.4772010021515822, 'm**3/kg'), Q_(1013250., 'J/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.vu[0], Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.vu[1], Q_(1013250., 'J/kg'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_sv(self):
        s = State(substance='water')
        s.sv = Q_(3028.9867985920914, 'J/(kg*K)'), Q_(0.4772010021515822, 'm**3/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.sv[0], Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.sv[1], Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_vs(self):
        s = State(substance='water')
        s.vs = Q_(0.4772010021515822, 'm**3/kg'), Q_(3028.9867985920914, 'J/(kg*K)')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.vs[0], Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.vs[1], Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_sh(self):
        s = State(substance='water')
        s.sh = Q_(3028.9867985920914, 'J/(kg*K)'), Q_(1061602.391543017, 'J/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.sh[0], Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.sh[1], Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_hs(self):
        s = State(substance='water')
        s.hs = Q_(1061602.391543017, 'J/kg'), Q_(3028.9867985920914, 'J/(kg*K)')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.hs[0], Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.hs[1], Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_vh(self):
        s = State(substance='water')
        s.vh = Q_(0.4772010021515822, 'm**3/kg'), Q_(1061602.391543017, 'J/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.vh[0], Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.vh[1], Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.u, Q_(1013250., 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))

    def test_set_hv(self):
        s = State(substance='water')
        s.hv = Q_(1061602.391543017, 'J/kg'), Q_(0.4772010021515822, 'm**3/kg')
        assert isclose_quant(s.T, Q_(373.1242958476843, 'K'))
        assert isclose_quant(s.p, Q_(101325., 'Pa'))
        assert isclose_quant(s.hv[0], Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.hv[1], Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.u, Q_(1013250, 'J/kg'))
        assert isclose_quant(s.s, Q_(3028.9867985920914, 'J/(kg*K)'))
        assert isclose_quant(s.v, Q_(0.4772010021515822, 'm**3/kg'))
        assert isclose_quant(s.h, Q_(1061602.391543017, 'J/kg'))
        assert isclose_quant(s.x, Q_(0.28475636946248034, 'dimensionless'))



from setuptools import setup, find_packages
import os.path as op
import sys

here = op.abspath(op.dirname(__file__))

version = {}
with open(op.join(here, 'thermostate', '_version.py'), mode='r') as version_file:
    exec(version_file.read(), version)

with open(op.join(here, 'README.md'), mode='r') as readme_file:
    readme = readme_file.read()

with open(op.join(here, 'CHANGELOG.md'), mode='r') as changelog_file:
    changelog = changelog_file.read()

long_description = readme + '\n\n' + changelog

install_requires = [
    'coolprop>=6.1.0,<6.3',
    'pint>=0.7.2,<0.10',
]

tests_require = [
    'pytest>=3.0.0',
    'pytest-cov>=2.3.1',
]

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
setup_requires = ['pytest-runner'] if needs_pytest else []

setup(
    name='thermostate',
    version=version['__version__'],
    description='A package to manage thermodynamic states',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/bryanwweber/thermostate',
    author='Bryan W. Weber',
    author_email='bryan.w.weber@gmail.com',
    license='BSD-3-clause',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows 8',
        'Operating System :: Microsoft :: Windows :: Windows 8.1',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
    ],
    packages=find_packages(),
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    python_requires='~=3.5',
)
